/*
 * Sample Implementation of Catalog for Java platform that will save data
 * on close().
 *
 * Rob Nielsen 01/06/1999
 * rnielsen@cygnus.uwa.edu.au
 * http://www.cygnus.uwa.edu.au/~rnielsen/
 */

package waba.io;

import java.io.*;
import waba.util.Vector;
import waba.sys.Vm;

/**
 * Catalog is a collection of records commonly referred to as a database
 * on small devices.
 * <p>
 * Here is an example showing data being read from records in a catalog:
 *
 * <pre>
 * Catalog c = new Catalog("CRTR.TYPE", Catalog.READ_ONLY);
 * if (!c.isOpen())
 *   return;
 * int count = c.getRecordCount();
 * byte b[] = new byte[10];
 * for (int i = 0; i < count; i++)
 *   {
 *   if (c.setRecord(i) != true)
 *     continue; // record is deleted
 *   c.readBytes(b, 0, 10);
 *   ...
 *   }
 * c.close();
 * </pre>
 */

public class Catalog extends Stream
{
  /** Read-only open mode. */
  public static final int READ_ONLY  = 1;
  /** Write-only open mode. */
  public static final int WRITE_ONLY = 2;
  /** Read-write open mode. */
  public static final int READ_WRITE = 3; // READ | WRITE
  /** Create open mode. Used to create a database if one does not exist. */
  public static final int CREATE = 4;
  /** Is the Catalog open? */
  private boolean open=false;

  /** The file for the Catalog */
  private java.io.File f;

  /** The records in the Catalog */
  private Vector records;

  /** The index of the currently selected record */
  private int currentIndex;

  /** The currently selected record */
  private byte[] currentRecord;

  /** The byte position in the current record */
  private int pos;

  /**
   * Opens a catalog with the given name and mode. If mode is CREATE, the
   * database will be created if it does not exist. The name must be exactly
   * 9 characters.
   * <p>
   * For PalmOS: The first 4 characters in the name specify the database's
   * creator. If you want the database to be deleted when your program is
   * deinstalled, it should have the same creator ID as your program's
   * creator ID. If you are using exegen, the creator ID is generated
   * automatically from your program name. When the exegen program is executed
   * it will display the creator id it generates. The 5th character must
   * be a period. The 6th through 9th characters specify the type of
   * database.
   * <p>
   * Here is an example:
   * <pre>
   * // creator is AbCD and type is TYPE
   * Catalog c = new Catalog("AbCD.TYPE", Catalog.CREATE);
   * </pre>
   * @param name 9 character database identifier
   * @param mode open/create mode
   */
  public Catalog(String name, int mode)
  {
    f=new java.io.File(name);
    if (f.exists()&&mode!=CREATE)
    {
      try
      {
        DataInputStream dis=new DataInputStream(new FileInputStream(f));
        int numRecords=dis.readInt();
        records=new Vector(numRecords);
        for(int i=0;i<numRecords;i++)
        {
          int len=dis.readInt();
          int numRead=0;
          byte[] buf=new byte[len];
          while (numRead<len)
          {
            numRead+=dis.read(buf,numRead,len-numRead);
          }
          records.add(buf);
        }
        dis.close();
        open=true;
      }
      catch (IOException e)
      {
        e.printStackTrace();
        open=false;
      }
    }
    else
    if (mode==CREATE)
    {
      records=new Vector();
      open=true;
    }
    else
      open=false;
    currentIndex=-1;
  }

  /**
   * Adds a record to the end of the catalog. If this operation is successful,
   * the position of the new record is returned and the current position is
   * set to the new record. If it is unsuccessful the current position is
   * unset and -1 is returned.
   * @param size the size in bytes of the record to add
   */
  public int addRecord(int size)
  {
    if (!open)
      return -1;
    byte[] buf=new byte[size];
    records.add(buf);
    int pos=records.getCount()-1;
    setRecordPos(pos);
    return pos;
  }

  /**
   * Closes the catalog. Returns true if the operation is successful and false
   * otherwise.
   */
  public boolean close()
  {
    if (!open)
      return false;
    else
      try
      {
        DataOutputStream dos=new DataOutputStream(new FileOutputStream(f));
        int size=records.getCount();
        dos.writeInt(size);
        for(int i=0;i<size;i++)
        {
          byte[] buf=(byte[])records.get(i);
          int len=buf.length;
          dos.writeInt(len);
          dos.write(buf,0,len);
        }
		dos.close();
        open=false;
        return true;
      }
      catch(IOException e)
      {
        e.printStackTrace();
        return false;
      }
  }

  /**
   * Deletes the catalog. Returns true if the operation is successful and false
   * otherwise. Under PalmOS, this operation simply marks the record for deletion.
   * The record will actually be removed by the operating system at some later
   * time. Under WindowsCE, the record is immediatly removed.
   */
  public boolean delete()
  {
    return f.delete();
  }

  /**
   * Deletes the current record and sets the current record position to -1.
   * A call to setRecordPos() for a deleted record will fail under PalmOS
   * until the operating system actually frees up the record.
   */
  public boolean deleteRecord()
  {
    if (!open)
      return false;
    records.del(currentIndex);
    currentIndex=-1;
    return true;
  }

  /**
   * Returns the number of records in the catalog or -1 if the catalog is not open.
   */
  public int getRecordCount()
  {
    return records.getCount();
  }

  /**
   * Returns the size of the current record in bytes or -1 if there is no
   * current record.
   */
  public int getRecordSize()
  {
    if (!open||currentIndex==-1)
      return -1;
    return currentRecord.length;
  }

  /**
   * Returns true if the catalog is open and false otherwise. This can
   * be used to check if opening or creating a catalog was successful.
   */
  public boolean isOpen()
  {
    return open;
  }

  /**
   * Sets the current record position and locks the given record. The value
   * -1 can be passed to unset and unlock the current record. If the operation
   * is succesful, true is returned and the read/write cursor is set to
   * the beginning of the record. Otherwise, false is returned (which
   * usually indicates the record has been deleted).
   */
  public boolean setRecordPos(int pos)
  {
    if (!open)
      return false;
    currentIndex=pos;
    if (pos!=-1)
    {
      currentRecord=(byte[])records.get(currentIndex);
      this.pos=0;
    }
    return true;
  }

  /**
   * Reads bytes from the current record into a byte array. Returns the
   * number of bytes actually read or -1 if an error prevented the
   * read operation from occurring. After the read is complete, the location of
   * the cursor in the current record (where read and write operations start from)
   * is advanced the number of bytes read.
   * @param buf the byte array to read data into
   * @param start the start position in the array
   * @param count the number of bytes to read
   */
  public int readBytes(byte buf[], int start, int count)
  {
    if (!open)
      return -1;
    Vm.copyArray(currentRecord,pos,buf,start,count);
    pos+=count;
    return count;
  }

  /**
   * Advances the cursor in the current record a number of bytes. The cursor
   * defines where read and write operations start from in the record. Returns
   * the number of bytes actually skipped or -1 if an error occurs.
   * @param count the number of bytes to skip
   */
  public int skipBytes(int count)
  {
    if (!open)
      return -1;
    pos+=count;
    return count;
  }

  /**
   * Writes to the current record. Returns the number of bytes written or -1
   * if an error prevented the write operation from occurring.
   * After the write is complete, the location of the cursor in the current record
   * (where read and write operations start from) is advanced the number of bytes
   * written.
   * @param buf the byte array to write data from
   * @param start the start position in the byte array
   * @param count the number of bytes to write
   */
  public int writeBytes(byte buf[], int start, int count)
  {
    if (!open)
      return -1;
    Vm.copyArray(buf,start,currentRecord,pos,count);
    pos+=count;
    return count;
  }
}